#include <stdio.h>
#include <qapplication.h>
#include <qlistbox.h>
#include <qlineedit.h>
#include <qfiledialog.h>
#include <qlabel.h>
#include <qdom.h>
#include <qtextbrowser.h>
#include "moc_pingvo.h"

IdxViewItem::IdxViewItem( int start, int end, char *text0 ):
	art_start(start), art_end(end), tstr(text0) {
}

QString IdxViewItem::text() const {
  return tstr;
}

IdxView::IdxView(): fname(QString::null), rootNode(0), xmlMode(false) {
  setFixedSize( size() );
  show();
}

IdxViewItem *IdxView::lookupNode( IdxViewItem *from,
	QString prefix, int pfxlen ) {
  IdxViewItem *n, *pnode;

  n = from;
  pnode = 0;
  while ( n ) {
  	if ( n->tstr.left( pfxlen ).lower() == prefix.left( pfxlen ) ) {
		if ( pfxlen < prefix.length() && n->firstChild ) {
			pfxlen++;
			pnode = n;
			n = n->firstChild;
		}
		else return n;
	}
	else n = n->nextSibling;
  }

  return pnode;
}

IdxViewItem *IdxView::lookupNode( QString prefix ) {
  if ( !rootNode ) return 0;
  return lookupNode( rootNode, prefix.lower(), 1 );
}

void IdxView::insertNode( IdxViewItem *node ) {
  IdxViewItem *n;

  node->firstChild = 0;
  n = lookupNode( node->tstr );
  if ( !n ) {
	if ( !rootNode ) {
		rootNode = node;
		node->prevSibling = 0;
		node->nextSibling = 0;
		return;
	}
	node->prevSibling = 0;
	node->nextSibling = rootNode;
	rootNode->prevSibling = node;
	rootNode = node;
  }

  if ( node->tstr.length() < n->tstr.length() ) {
  	if ( rootNode == n ) rootNode = node;
	node->prevSibling = n->prevSibling;
	if ( node->prevSibling )
		node->prevSibling->nextSibling = node;
	node->nextSibling = n->nextSibling;
	node->firstChild = n;
	n->prevSibling = node;
	n->nextSibling = 0;
  }
  else {
  	node->prevSibling = n;
  	node->nextSibling = n->firstChild;
	n->firstChild = node;
  }
  dumpTree();
}

void IdxView::dumpChilds( IdxViewItem *from, int indent ) {
  int i;
  IdxViewItem* node = from->firstChild;

  while ( node ) {
	for ( i = 0; i < indent; i++ ) printf( " " );
	printf( "%s\n", node->tstr.latin1() );
	dumpChilds( node, indent+1 );
	node = node->nextSibling;
  }
}

void IdxView::dumpTree() {
  if ( !rootNode ) {
  	puts( "(empty tree)" );
  	return;
  }

  puts( "----" );
  printf( "%s\n", rootNode->tstr.latin1() );
  dumpChilds( rootNode, 1 );
  puts( "----" );
}

void IdxView::loadIndex( char *prefix ) {
  FILE *f;
  int pos, rc, pfxlen;
  char *buf;
  IdxViewItem *node, *prev;

  if ( prefix ) pfxlen = strlen( prefix );
  f = fopen( fname.latin1(), "r" );
  if ( !f ) {
  	perror( "fopen" );
  	return;
  }

  prev = 0;
  while ( !feof(f) ) {
	rc = fscanf( f, "%d %a[^\n]", &pos, &buf );
	if ( rc != 2 ) break;
	if ( prefix && *prefix && strncasecmp( buf, prefix, pfxlen ) ) {
		free( buf );
		continue;
	}
	node = new IdxViewItem( pos, -1, buf );
	lb->insertItem( node );
//	insertNode( node );
	free( buf );
	if ( prev ) prev->art_end = pos;
	prev = node;
  }
  fclose( f );
//  dumpTree();
}

void IdxView::loadIndexXML( char *prefix ) {
  QFile f( fname );
  QDomDocument doc("dictionary");
  QDomElement docEl, e;
  QDomNode n;
  IdxViewItem *node;
  QString pfx = QString(prefix).lower();

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open dictionary" );
	return;
  }
  doc.setContent( &f );
  f.close();
  docEl = doc.documentElement();
  for ( n = docEl.firstChild(); !n.isNull(); n = n.nextSibling() ) {
	if ( !n.isElement() ) continue;
	e = n.toElement();
	if ( e.tagName() != "entry" ||
	     e.attribute("word").left(pfx.length()).lower() != pfx ) continue;
	node = new IdxViewItem( -1, -1, (char*) e.attribute("word").latin1() );
	lb->insertItem( node );
	
  }
}

void IdxView::lookupByPrefix() {
  lb->clear();
  if ( xmlMode )
  	loadIndexXML( (char*) le->text().latin1() );
  else
	loadIndex( (char*) le->text().latin1() );
  lb->setSelected( lb->firstItem(), true );
}

void IdxView::loadEnRu() {
  lb->clear();
  fname = "lingvo-universal-en-ru.idx";
  dfname = "lingvo-universal-en-ru.dict";
  loadIndex( (char*) le->text().latin1() );
  lb->setSelected( lb->firstItem(), true );
  xmlMode = false;
}

void IdxView::loadRuEn() {
  lb->clear();
  fname = "";
  dfname = "";
  xmlMode = false;
}

void IdxView::loadDefXML() {
  lb->clear();
  fname = "lingvo.xml";
  dfname = "";
  loadIndexXML( (char*) le->text().latin1() );
  xmlMode = true;
}

void IdxView::loadCustom() {
  QString s = QFileDialog::getOpenFileName();
  if ( s.isEmpty() ) return;
  lb->clear();
  fname = s;
  if ( fname.right(4) == ".idx" ) {
  	dfname = fname.left( fname.length() - 4 ) + ".dict";
	loadIndex( (char*) le->text().latin1() );
	xmlMode = false;
  }
  else if ( fname.right(4).lower() == ".xml" ) {
	dfname = "";
  	loadIndexXML( (char*) le->text().latin1() );
	xmlMode = true;
  }
  lb->setSelected( lb->firstItem(), true );
}

QDomElement IdxView::getXMLArticle( QString word ) {
  QFile f( fname );
  QDomDocument doc("dictionary");
  QDomElement docEl, e;
  QDomNode n;

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open dictionary" );
	return QDomElement();
  }
  doc.setContent( &f );
  f.close();
  docEl = doc.documentElement();
  for ( n = docEl.firstChild(); !n.isNull(); n = n.nextSibling() ) {
	if ( !n.isElement() ) continue;
	e = n.toElement();
	if ( e.tagName() == "entry" && e.attribute("word") == word ) return e;
	
  }

  return QDomElement();
}

void IdxView::showArticle( QListBoxItem *lbItem ) {
  ArticleView *av;
  IdxViewItem *node;

  node = (IdxViewItem*) lbItem;
//  av->loadArticle();
  if ( xmlMode ) {
  	QDomElement e = getXMLArticle( node->text() );
	if ( e.isNull() ) return;
  	av = new ArticleView();
	av->loadArticleXML( e );
  }
}

ArticleView::ArticleView() {
//  enableClipper( TRUE );
//  viewport()->setEraseColor( QColor( 0xff, 0xff, 0xff ) );
//  show();
}

void ArticleView::loadArticle( char *fname, int start, int stop ) {
#if 0
  FILE *f;
  int x, y, rc;
  char *text;
  QLabel *l;

  f = fopen( fname, "r" );
  if ( !f ) {
	perror( "fopen" );
	return;
  }
  fseek( f, start, SEEK_SET );
  while ( !feof(f) && (stop == -1 || ftell(f) < stop) ) {
	rc = fscanf( f, "%d %d {%a[^}]}\n", &x, &y, &text );
	if ( rc != 3 ) break;
	l = new QLabel( QString::fromLocal8Bit(text), viewport() );
	l->setEraseColor( QColor( 0xff, 0xff, 0xff ) );
	l->adjustSize();
	addChild( l, x, y );
  }
  fclose( f );

  QRect r = viewport()->childrenRect();
  resizeContents( r.width(), r.height() );
#endif
}

QString ArticleView::transXMLtoRichText( QDomElement e ) {
  QDomNode n;
  QString text;
  
  for ( n = e.firstChild(); !n.isNull(); n = n.nextSibling() ) {
  	if ( n.isText() )
		text += n.toText().data();
	else if ( n.isElement() ) {
		e = n.toElement();
#if 0
		if ( e.tagName() == "m1" )
			text += "<p>&nbsp;" + transXMLtoRichText( e ) +
				"</p>";
		else if ( e.tagName() == "m2" )
			text += "<p>&nbsp;&nbsp;" + transXMLtoRichText( e ) +
				"</p>";
#else
		if ( e.tagName() == "m1" )
			text += "<tr><td>&nbsp;" + transXMLtoRichText( e ) +
				"</td></tr>";
		else if ( e.tagName() == "m2" )
			text += "<tr><td>&nbsp;&nbsp;" + transXMLtoRichText( e ) +
				"</td></tr>";
#endif
		else if ( e.tagName() == "com" )
			text += "<i><font color=\"#007F00\">" +
				transXMLtoRichText( e ) + "</font></i>\n";
		else if ( e.tagName() == "ex" )
			text += "<font color=\"#7F7F7F\">" +
				transXMLtoRichText( e ) + "</font>\n";
		else if ( e.tagName() == "trn" )
			text += transXMLtoRichText( e );
		else if ( e.tagName() == "b" )
			text += "<b>" + transXMLtoRichText( e ) +
				"</b>";
		else if ( e.tagName() == "ref" ) {
			QString ref = transXMLtoRichText( e );
			text += "<a href=\"#\">" + ref + "</a>";
		}
		else
			printf( "element with tag: %s\n", e.tagName().latin1() );
	}
  }

  return text;
}

void ArticleView::loadArticleXML( QDomElement e ) {
  QString text = "<p><h2>" + e.attribute("word") + "</h2></p>\n" +
  	"<table border=0 cellspacing=0 cellpadding=0>" +
	transXMLtoRichText( e ) + "</table>";

  QTextBrowser *tb = new QTextBrowser;
  tb->setText( text );
  tb->resize( 420, 470 );
  tb->show();

//  QLabel *l = new QLabel( text, viewport() );
//  l->setTextFormat( RichText );
//  l->setEraseColor( QColor( 0xff, 0xff, 0xff ) );
//  l->adjustSize();
//  addChild( l, 0, 0 );

//  QRect r = viewport()->childrenRect();
//  resizeContents( r.width(), r.height() );
}

int main( int argc, char **argv ) {
  QApplication qa( argc, argv );

  IdxView w;
  qa.setMainWidget( &w );

  return qa.exec();
}
